from typing import override, ClassVar, Type

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = [
    "ConvNeXt",
    "ConvNeXt_V2",
    "convnext_tiny",
    "convnext_small",
    "convnext_base",
    "convnext_large",
    "convnext_xlarge",
    "convnext_v2_atto",
    "convnext_v2_femto",
    "convnext_v2_pico",
    "convnext_v2_nano",
    "convnext_v2_tiny",
    "convnext_v2_base",
    "convnext_v2_large",
    "convnext_v2_huge",
]


class _Block(nn.Module):
    def __init__(
        self, in_channels: int, drop_path: float = 0.0, layer_scale_init: float = 1e-6
    ) -> None:
        super().__init__()

        self.dwconv = nn.Conv2d(
            in_channels, in_channels, kernel_size=7, padding=3, groups=in_channels
        )
        self.norm = nn.LayerNorm(in_channels, eps=1e-6)

        self.pwconv1 = nn.Linear(in_channels, 4 * in_channels)
        self.gelu = nn.GELU()
        self.pwconv2 = nn.Linear(4 * in_channels, in_channels)

        self.gamma = (
            nn.Parameter(layer_scale_init * lucid.ones(in_channels))
            if layer_scale_init > 0
            else None
        )
        self.drop_path = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        input_ = x
        x = self.dwconv(x)
        x = x.transpose((0, 2, 3, 1))
        x = self.norm(x)

        n, h, w, _ = x.shape
        x = self.pwconv1(x.reshape(n * h * w, -1))
        x = self.gelu(x)
        x = self.pwconv2(x)

        if self.gamma is not None:
            flat_gamma = lucid.repeat(self.gamma, x.shape[0], axis=0)
            x = flat_gamma * x.reshape(-1)

        x = x.reshape(n, -1, h, w)
        x = input_ + self.drop_path(x)
        return x


class _ChannelsFisrtLayerNorm(nn.LayerNorm):
    def __init__(
        self,
        normalized_shape: int,
        eps: float = 0.00001,
        elementwise_affine: bool = True,
        bias: bool = True,
    ) -> None:
        super().__init__(normalized_shape, eps, elementwise_affine, bias)

    @override
    def forward(self, x: Tensor) -> Tensor:
        x = x.transpose((0, 2, 3, 1))
        out = super().forward(x)
        out = out.transpose((0, 3, 1, 2))
        return out


class ConvNeXt(nn.Module):
    base_block: ClassVar[Type[nn.Module]] = _Block

    def __init__(
        self,
        num_classes: int = 1000,
        depths: list[int] = [3, 3, 9, 3],
        dims: list[int] = [96, 192, 384, 768],
        drop_path: float = 0.0,
        layer_scale_init: float = 1e-6,
    ) -> None:
        super().__init__()

        self.downsample_layers = nn.ModuleList()
        stem = nn.Sequential(
            nn.Conv2d(3, dims[0], kernel_size=4, stride=4),
            _ChannelsFisrtLayerNorm(dims[0]),
        )
        self.downsample_layers.append(stem)

        for i in range(3):
            downsample = nn.Sequential(
                _ChannelsFisrtLayerNorm(dims[i]),
                nn.Conv2d(dims[i], dims[i + 1], kernel_size=2, stride=2),
            )
            self.downsample_layers.append(downsample)

        self.stages = nn.ModuleList()
        dp_rates = [x.item() for x in lucid.linspace(0, drop_path, sum(depths))]
        cur = 0
        for i in range(4):
            stage = nn.Sequential(
                *[
                    self.base_block(dims[i], dp_rates[cur + j], layer_scale_init)
                    for j in range(depths[i])
                ]
            )
            self.stages.append(stage)
            cur += depths[i]

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.norm = nn.LayerNorm(dims[-1])
        self.head = nn.Linear(dims[-1], num_classes)

    def forward(self, x: Tensor) -> Tensor:
        for i in range(4):
            x = self.downsample_layers[i](x)
            x = self.stages[i](x)

        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)

        x = self.norm(x)
        x = self.head(x)

        return x


class _Block_V2(nn.Module):
    def __init__(self, channels: int, drop_path: float = 0.0, *args) -> None:
        super().__init__()
        self.dwconv = nn.Conv2d(
            channels, channels, kernel_size=7, padding=3, groups=channels
        )
        self.norm = nn.LayerNorm(channels)

        self.pwconv1 = nn.Linear(channels, 4 * channels)
        self.gelu = nn.GELU()
        self.grn = nn.GlobalResponseNorm(4 * channels)

        self.pwconv2 = nn.Linear(4 * channels, channels)
        self.drop_path = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        input_ = x
        x = self.dwconv(x).transpose((0, 2, 3, 1))
        x = self.norm(x)

        n, h, w, _ = x.shape
        x = self.pwconv1(x.reshape(n * h * w, -1))
        x = self.gelu(x)
        x = self.grn(x.reshape(n, -1, h, w))
        x = self.pwconv2(x.reshape(n * h * w, -1))

        x = x.reshape(n, -1, h, w)
        x = input_ + self.drop_path(x)

        return x


class ConvNeXt_V2(ConvNeXt):
    base_block: ClassVar[Type[nn.Module]] = _Block_V2

    def __init__(
        self,
        num_classes: int = 1000,
        depths: list[int] = [3, 3, 9, 3],
        dims: list[int] = [96, 192, 384, 768],
        drop_path: float = 0.0,
    ) -> None:
        super().__init__(num_classes, depths, dims, drop_path)


@register_model
def convnext_tiny(num_classes: int = 1000, **kwargs) -> ConvNeXt:
    depths = [3, 3, 9, 3]
    dims = [96, 192, 384, 768]
    return ConvNeXt(num_classes, depths, dims, **kwargs)


@register_model
def convnext_small(num_classes: int = 1000, **kwargs) -> ConvNeXt:
    depths = [3, 3, 27, 3]
    dims = [96, 192, 364, 768]
    return ConvNeXt(num_classes, depths, dims, **kwargs)


@register_model
def convnext_base(num_classes: int = 1000, **kwargs) -> ConvNeXt:
    depths = [3, 3, 27, 3]
    dims = [128, 256, 512, 1024]
    return ConvNeXt(num_classes, depths, dims, **kwargs)


@register_model
def convnext_large(num_classes: int = 1000, **kwargs) -> ConvNeXt:
    depths = [3, 3, 27, 3]
    dims = [192, 384, 768, 1536]
    return ConvNeXt(num_classes, depths, dims, **kwargs)


@register_model
def convnext_xlarge(num_classes: int = 1000, **kwargs) -> ConvNeXt:
    depths = [3, 3, 27, 3]
    dims = [256, 512, 1024, 2048]
    return ConvNeXt(num_classes, depths, dims, **kwargs)


@register_model
def convnext_v2_atto(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    depths = [2, 2, 6, 2]
    dims = [40, 80, 160, 320]
    return ConvNeXt_V2(num_classes, depths, dims, **kwargs)


@register_model
def convnext_v2_femto(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    depths = [2, 2, 6, 2]
    dims = [48, 96, 192, 384]
    return ConvNeXt_V2(num_classes, depths, dims, **kwargs)


@register_model
def convnext_v2_pico(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    depths = [2, 2, 6, 2]
    dims = [64, 128, 256, 512]
    return ConvNeXt_V2(num_classes, depths, dims, **kwargs)


@register_model
def convnext_v2_nano(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    depths = [2, 2, 8, 2]
    dims = [80, 160, 320, 640]
    return ConvNeXt_V2(num_classes, depths, dims, **kwargs)


@register_model
def convnext_v2_tiny(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    depths = [3, 3, 9, 3]
    dims = [96, 192, 384, 768]
    return ConvNeXt_V2(num_classes, depths, dims, **kwargs)


@register_model
def convnext_v2_base(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    depths = [3, 3, 27, 3]
    dims = [128, 256, 512, 1024]
    return ConvNeXt_V2(num_classes, depths, dims, **kwargs)


@register_model
def convnext_v2_large(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    depths = [3, 3, 27, 3]
    dims = [192, 384, 768, 1536]
    return ConvNeXt_V2(num_classes, depths, dims, **kwargs)


@register_model
def convnext_v2_huge(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    depths = [3, 3, 27, 3]
    dims = [352, 704, 1408, 2816]
    return ConvNeXt_V2(num_classes, depths, dims, **kwargs)



"""
# `Lucid `

**Lucid** is an educational deep learning framework developed to help users understand
the underlying mechanics of deep learning models and tensor operations.

It is designed to provide a simple yet powerful environment to experiment with neural networks,
optimization, and backpropagation using only `NumPy`.

Lucid is ideal for those who want to learn about the inner workings of deep learning
algorithms and operations without the complexity of high-level frameworks.

[📑 Lucid Documentation](https://chanlumerico.github.io/lucid/build/html/index.html)
"""

from contextlib import contextmanager, AbstractContextManager
from typing import Any, Generator, SupportsIndex, Callable, Self, Optional, Type
from types import TracebackType
from functools import wraps
from pathlib import Path

import os
import sys
import json
import math
import numpy as np

from lucid._tensor import Tensor
from lucid._func import *
from lucid._util import *

from lucid._backend.metal import mx

from lucid.types import (
    _ArrayOrScalar,
    _NumPyArray,
    _MLXArray,
    _ArrayLike,
    _ShapeLike,
    _DeviceType,
    _BuiltinNumeric,
    Numeric,
)
from lucid.error import *

import lucid.linalg as linalg
import lucid.random as random
import lucid.einops as einops
import lucid.nn as nn
import lucid.types as types


_grad_enabled: bool = True
_flops_enabled: bool = False

newaxis = None

pi = math.pi
inf = math.inf

Int = types.Int
Int8, Int16, Int32, Int64 = (types.Int8, types.Int16, types.Int32, types.Int64)

Float = types.Float
Float16, Float32, Float64 = (types.Float16, types.Float32, types.Float64)

Complex = types.Complex
Complex64 = types.Complex64


def tensor(
    data: Tensor | _ArrayOrScalar,
    requires_grad: bool = False,
    keep_grad: bool = False,
    dtype: _BuiltinNumeric | Numeric | None = None,
) -> Tensor:
    if isinstance(data, Tensor):
        data = data.data
    return Tensor(data, requires_grad, keep_grad, dtype)


def to_tensor(
    a: _ArrayLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    dtype: _BuiltinNumeric | Numeric | None = None,
) -> Tensor:
    return tensor(a, requires_grad, keep_grad, dtype)


class _NoGrad(AbstractContextManager):
    __slots__ = ("_prev_state",)

    def __enter__(self) -> Self:
        global _grad_enabled
        self._prev_state = _grad_enabled

        _grad_enabled = False
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> bool:
        _ = (exc_type, exc_value, traceback)
        global _grad_enabled

        _grad_enabled = self._prev_state
        return False

    def __call__(self, func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            with _NoGrad():
                return func(*args, **kwargs)

        return wrapper


no_grad = _NoGrad


def grad_enabled() -> bool:
    return _grad_enabled


@contextmanager
def count_flops() -> Generator:
    global _flops_enabled
    prev_state = _flops_enabled
    _flops_enabled = True
    try:
        yield
    finally:
        _flops_enabled = prev_state


def flops_enabled() -> bool:
    return _flops_enabled


def shape(a: Tensor | _NumPyArray | _MLXArray) -> _ShapeLike:
    if hasattr(a, "shape"):
        return a.shape
    raise ValueError(f"The argument must be a Tensor or a NumPy array.")


def _check_input_dim(tensor: Tensor, dim: int) -> None:
    if tensor.ndim != dim:
        raise ValueError(f"expected {dim}D input (got {tensor.ndim}D input).")


def _set_tensor_grad(
    tensor: Tensor, grad: _NumPyArray | _MLXArray, at: SupportsIndex = ...
) -> None:
    if not tensor.requires_grad:
        return
    if tensor.grad is None:
        tensor.grad = grad
    else:
        if tensor.is_cpu() and not tensor.grad.flags.writeable:
            tensor.grad = tensor.grad.copy()

        if tensor.is_gpu():
            if at == Ellipsis:
                at = slice(None, None, None)

        if tensor.grad.ndim == 0:
            tensor.grad += grad
        else:
            tensor.grad[at] = tensor.grad[at] + grad


def _check_is_tensor(
    any: Tensor | _ArrayOrScalar, device: _DeviceType = "cpu"
) -> Tensor:
    if not isinstance(any, Tensor):
        return Tensor(any, device=device)
    return any


def _match_grad_shape(
    data: _NumPyArray | _MLXArray,
    grad: _NumPyArray | _MLXArray,
    device: _DeviceType = "cpu",
) -> _NumPyArray | _MLXArray:
    if data.shape == grad.shape:
        return grad
    if data.ndim == 0:
        return grad.sum()
    if grad.ndim == 0:
        return (
            np.broadcast_to(grad, data.shape)
            if device == "cpu"
            else mx.broadcast_to(grad, data.shape)
        )

    if data.size == grad.size:
        return grad.reshape(data.shape)

    elif data.size > grad.size:
        grad_squeeze = grad.flatten()
        expand_factor = data.size / grad.size
        if expand_factor % 1 != 0:
            raise ValueError(
                f"Cannot broadcast grad of {grad.shape} to data of {data.shape}."
            )
        grad_expand = (
            grad_squeeze[..., None].repeat(int(expand_factor), axis=-1)
            if device == "cpu"
            else mx.repeat(grad_squeeze[..., None], int(expand_factor), axis=1)
        )
        return grad_expand.reshape(data.shape)

    elif data.size < grad.size:
        if grad.size % data.size != 0:
            raise ValueError(
                f"Cannot collapse grad of {grad.shape} to data of {data.shape}."
            )
        new_shape = tuple()
        remain_size = grad.size

        for d_dim in data.shape:
            fac = remain_size // d_dim
            new_shape += (d_dim,)
            remain_size = fac

        new_shape += (fac,)
        return grad.reshape(new_shape).sum(axis=-1)

    else:
        raise ValueError("Unknown error occurred.")


def _get_overloaded_shape(args: int | _ShapeLike) -> _ShapeLike:
    if len(args) == 1 and isinstance(args[0], (tuple, list)):
        shape = tuple(args[0])
    else:
        shape = tuple(args)
    return shape


REGISTRY_PATH: Path = Path("lucid/models/registry.json")

_ModuleReturnFunc = Callable[[Any], nn.Module]


def register_model(func: _ModuleReturnFunc) -> _ModuleReturnFunc:
    @wraps(func)
    def wrapper(*args, **kwargs) -> nn.Module:
        if os.environ.get("SPHINX_BUILD"):
            return func(*args, **kwargs)

        if not REGISTRY_PATH.exists():
            REGISTRY_PATH.parent.mkdir(parents=True, exist_ok=True)
            with open(REGISTRY_PATH, "w") as f:
                json.dump({}, f)

        with open(REGISTRY_PATH, "r") as f:
            registry = json.load(f)

        model = func(*args, **kwargs)
        model._alt_name = func.__name__
        name = func.__name__

        if name in registry:
            return model

        family = model.__class__.__name__
        param_size = model.parameter_size
        arch = sys.modules[func.__module__].__package__.replace("lucid.models.", "")

        registry[name] = dict(
            name=name,
            family=family,
            param_size=param_size,
            arch=arch,
        )

        with open(REGISTRY_PATH, "w") as f:
            json.dump(registry, f, indent=4)

        return model

    return wrapper



from abc import ABC, abstractmethod
from typing import Self, Any
import random
import math


class Dataset(ABC):
    @abstractmethod
    def __getitem__(self, index: int) -> None:
        raise NotImplementedError("Subclasses must implement __getitem__.")

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError("Subclasses must implement __len__.")

    def __add__(self, other: Self) -> Self:
        return ConcatDataset([self, other])


class ConcatDataset(Dataset):
    def __init__(self, datasets: list[Dataset]) -> None:
        super().__init__()
        self.datasets = datasets
        self.cumulative_sizes = self._compute_cumulative_sizes()

    def _compute_cumulative_sizes(self) -> list[int]:
        cum_sizes = []
        total = 0
        for dataset in self.datasets:
            total += len(dataset)
            cum_sizes.append(total)

        return cum_sizes

    def __len__(self) -> int:
        return self.cumulative_sizes[-1] if self.cumulative_sizes else 0

    def __getitem__(self, idx: int) -> Any:
        if idx < 0:
            if -idx > len(self):
                raise IndexError("Index out of range.")
            idx = len(self) + idx

        for i, size in enumerate(self.cumulative_sizes):
            if idx < size:
                dataset_idx = i
                if dataset_idx > 0:
                    idx -= self.cumulative_sizes[dataset_idx - 1]

                return self.datasets[dataset_idx][idx]

        raise IndexError("Index out of range.")


class DataLoader:
    def __init__(
        self, dataset: Dataset, batch_size: int = 1, shuffle: bool = False
    ) -> None:
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.indices = list(range(len(dataset)))

        if shuffle:
            self._shuffle_indices()

    def _shuffle_indices(self) -> None:
        random.shuffle(self.indices)

    def __iter__(self) -> Self:
        self.current_index = 0
        if self.shuffle:
            self._shuffle_indices()
        return self

    def __next__(self):
        if self.current_index >= len(self.indices):
            raise StopIteration

        start = self.current_index
        end = min(start + self.batch_size, len(self.indices))

        batch_indices = self.indices[start:end]
        batch = self.dataset[batch_indices]

        self.current_index = end
        return batch

    def __len__(self) -> int:
        return int(math.ceil(len(self.dataset) / self.batch_size))



import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = ["Xception", "xception"]


class _Block(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        reps: int,
        stride: int = 1,
        start_with_relu: bool = True,
        grow_first: bool = True,
    ) -> None:
        super().__init__()
        if out_channels != in_channels or stride != 1:
            self.skip = nn.Conv2d(
                in_channels, out_channels, kernel_size=1, stride=stride, bias=False
            )
            self.skipbn = nn.BatchNorm2d(out_channels)
        else:
            self.skip = None

        rep = []
        channels = in_channels
        if grow_first:
            rep.append(nn.ReLU())
            rep.append(
                nn.DepthSeparableConv2d(
                    in_channels, out_channels, kernel_size=3, padding=1, bias=False
                )
            )
            rep.append(nn.BatchNorm2d(out_channels))
            channels = out_channels

        for i in range(reps - 1):
            rep.append(nn.ReLU())
            rep.append(
                nn.DepthSeparableConv2d(
                    channels, channels, kernel_size=3, padding=1, bias=False
                )
            )
            rep.append(nn.BatchNorm2d(channels))

        if not grow_first:
            rep.append(nn.ReLU())
            rep.append(
                nn.DepthSeparableConv2d(
                    in_channels, out_channels, kernel_size=3, padding=1, bias=False
                )
            )
            rep.append(nn.BatchNorm2d(out_channels))

        if not start_with_relu:
            rep = rep[1:]
        else:
            rep[0] = nn.ReLU()

        if stride != 1:
            rep.append(nn.MaxPool2d(kernel_size=3, stride=stride, padding=1))

        self.rep = nn.Sequential(*rep)

    def forward(self, x: Tensor) -> Tensor:
        out = self.rep(x)

        if self.skip is not None:
            skip = self.skip(x)
            skip = self.skipbn(skip)
        else:
            skip = x

        out += skip
        return out


class Xception(nn.Module):
    def __init__(self, num_classes: int = 1000) -> None:
        super().__init__()
        self.relu = nn.ReLU()

        self.conv1 = nn.ConvBNReLU2d(
            3, 32, kernel_size=3, stride=2, padding=0, conv_bias=False
        )
        self.conv2 = nn.ConvBNReLU2d(32, 64, kernel_size=3, conv_bias=False)

        self.block1 = _Block(64, 128, reps=2, stride=2, start_with_relu=False)
        self.block2 = _Block(128, 256, reps=2, stride=2)
        self.block3 = _Block(256, 728, reps=2, stride=2)

        self.mid_blocks = nn.Sequential(*[_Block(728, 728, reps=3) for _ in range(8)])
        self.end_block = _Block(728, 1024, reps=2, stride=2, grow_first=False)

        self.conv3 = nn.DepthSeparableConv2d(1024, 1536, kernel_size=3, padding=1)
        self.bn3 = nn.BatchNorm2d(1536)

        self.conv4 = nn.DepthSeparableConv2d(1536, 2048, kernel_size=3, padding=1)
        self.bn4 = nn.BatchNorm2d(2048)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(2048, num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.conv2(x)

        x = self.block3(self.block2(self.block1(x)))
        x = self.mid_blocks(x)
        x = self.end_block(x)

        x = self.relu(self.bn3(self.conv3(x)))
        x = self.relu(self.bn4(self.conv4(x)))

        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)

        return x


@register_model
def xception(num_classes: int = 1000, **kwargs) -> Xception:
    return Xception(num_classes, **kwargs)



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor

from typing import Literal, override


__all__ = ["MSELoss", "BCELoss", "CrossEntropyLoss", "NLLLoss", "HuberLoss"]


_ReductionType = Literal["mean", "sum"]


class _Loss(nn.Module):
    def __init__(self, reduction: _ReductionType | None = "mean") -> None:
        super().__init__()
        if reduction not in {"mean", "sum"}:
            raise ValueError(f"Got an unexpected reduction type: {reduction}.")
        self.reduction = reduction

    @override
    def forward(self, input_: Tensor, target: Tensor) -> Tensor:
        NotImplemented


class _WeightedLoss(nn.Module):
    def __init__(
        self, weight: Tensor | None = None, reduction: _ReductionType | None = "mean"
    ) -> None:
        super().__init__()
        if reduction not in {"mean", "sum", None}:
            raise ValueError(f"Got an unexpected reduction type: {reduction}.")
        self.reduction = reduction

        self.weight: Tensor | None
        self.register_buffer("weight", weight)

    @override
    def forward(self, input_: Tensor, target: Tensor) -> Tensor:
        NotImplemented


class MSELoss(_Loss):
    def forward(self, input_: Tensor, target: Tensor) -> Tensor:
        return F.mse_loss(input_, target, reduction=self.reduction)


class BCELoss(_WeightedLoss):
    def __init__(
        self,
        weight: Tensor | None = None,
        reduction: _ReductionType | None = "mean",
        eps: float = 1e-7,
    ) -> None:
        super().__init__(weight, reduction)
        self.eps = eps

    def forward(self, input_: Tensor, target: Tensor) -> Tensor:
        return F.binary_cross_entropy(
            input_, target, weight=self.weight, reduction=self.reduction, eps=self.eps
        )


class CrossEntropyLoss(_WeightedLoss):
    def __init__(
        self,
        weight: Tensor | None = None,
        reduction: _ReductionType | None = "mean",
        eps: float = 1e-7,
    ) -> None:
        super().__init__(weight, reduction)
        self.eps = eps

    def forward(self, input_: Tensor, target: Tensor) -> Tensor:
        return F.cross_entropy(
            input_, target, weight=self.weight, reduction=self.reduction, eps=self.eps
        )


class NLLLoss(_WeightedLoss):
    def forward(self, input_: Tensor, target: Tensor) -> Tensor:
        return F.nll_loss(input_, target, weight=self.weight, reduction=self.reduction)


class HuberLoss(_Loss):
    def __init__(
        self, reduction: _ReductionType | None = "mean", delta: float = 1.0
    ) -> None:
        super().__init__(reduction)
        self.delta = delta

    def forward(self, input_: Tensor, target: Tensor) -> Tensor:
        return F.huber_loss(input_, target, delta=self.delta, reduction=self.reduction)



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor


# NOTE: WIP
class RegionWarper(nn.Module):
    def __init__(self, output_size: tuple[int, int] = (224, 224)) -> None:
        super().__init__()
        self.output_size = output_size

    def forward(self, images: Tensor, rois: list[Tensor]) -> Tensor: ...



import lucid
import lucid.nn as nn
import lucid.nn.functional as F
from lucid._tensor import Tensor
import lucid.nn.functional


__all__ = [
    "Normalize",
    "Resize",
    "RandomHorizontalFlip",
    "RandomVerticalFlip",
    "RandomCrop",
    "CenterCrop",
    "RandomRotation",
    "RandomGrayscale",
]


class Normalize(nn.Module):
    def __init__(self, mean: tuple[float, ...], std: tuple[float, ...]) -> None:
        super().__init__()
        self.mean = lucid.tensor(mean)
        self.std = lucid.tensor(std)

    def forward(self, img: Tensor) -> Tensor:
        return (img - self.mean) / self.std


class Resize(nn.Module):
    def __init__(self, size: tuple[int, int]) -> None:
        super().__init__()
        self.size = size

    def forward(self, img: Tensor) -> Tensor:
        return F.interpolate(img, size=self.size, mode="bilinear")


class RandomHorizontalFlip(nn.Module):
    def __init__(self, p: float = 0.5) -> None:
        super().__init__()
        self.p = p

    def forward(self, img: Tensor) -> Tensor:
        if lucid.random.uniform().item() < self.p:
            return img[:, :, :, ::-1]
        return img


class RandomVerticalFlip(nn.Module):
    def __init__(self, p: float = 0.5) -> None:
        super().__init__()
        self.p = p

    def forward(self, img: Tensor) -> Tensor:
        if lucid.random.uniform().item() < self.p:
            return img[:, :, ::-1, :]
        return img


class RandomCrop(nn.Module):
    def __init__(self, size: tuple[int, int]) -> None:
        super().__init__()
        self.size = size

    def forward(self, img: Tensor) -> Tensor:
        _, _, H, W = img.shape
        crop_h, crop_w = self.size
        top = lucid.random.randint(0, H - crop_h + 1)
        left = lucid.random.randint(0, W - crop_w + 1)

        top = top.astype(int).squeeze().item()
        left = left.astype(int).squeeze().item()

        return img[:, :, top : top + crop_h, left : left + crop_w]


class CenterCrop(nn.Module):
    def __init__(self, size: tuple[int, int]) -> None:
        super().__init__()
        self.size = size

    def forward(self, img: Tensor) -> Tensor:
        _, _, H, W = img.shape
        crop_h, crop_w = self.size
        top = (H - crop_h) // 2
        left = (W - crop_w) // 2

        return img[:, :, top : top + crop_h, left : left + crop_w]


class RandomRotation(nn.Module):
    def __init__(self, degrees: float) -> None:
        super().__init__()
        self.degrees = degrees

    def forward(self, img: Tensor) -> Tensor:
        angle = lucid.random.uniform(-self.degrees, self.degrees)
        return lucid.nn.functional.rotate(img, angle.squeeze().item())


class RandomGrayscale(nn.Module):
    def __init__(self, p: float = 0.1):
        super().__init__()
        self.p = p

    def forward(self, img: Tensor) -> Tensor:
        if lucid.random.uniform() < self.p:
            r, g, b = img[:, 0:1, ...], img[:, 1:2, ...], img[:, 2:3, ...]
            grayscale = 0.299 * r + 0.587 * g + 0.114 * b
            img = grayscale.repeat(3, axis=1)
        return img



from collections import defaultdict
from typing import Any, Iterable
from abc import ABC, abstractmethod
import copy

import lucid.nn as nn

from lucid.types import _StateDict, _OptimClosure


class Optimizer(ABC):
    def __init__(
        self, params: Iterable[nn.Parameter], defaults: dict[str, Any]
    ) -> None:
        super().__init__()
        if not isinstance(params, Iterable):
            raise TypeError("params should be an iterable of Parameters.")

        params = list(params)
        self.param_groups = self.param_groups_setup(params, defaults)
        self.defaults = defaults
        self.state: dict[nn.Parameter, dict[str, Any]] = defaultdict(dict)

    def param_groups_setup(
        self, params: list[nn.Parameter], defaults: dict[str, Any]
    ) -> list[dict[str, Any]]:
        return [{"params": list(params), **defaults}]

    @abstractmethod
    def step(self, closure: _OptimClosure | None = None) -> Any | None:
        raise NotImplementedError("The step method must be implemented by subclasses.")

    def zero_grad(self) -> None:
        for group in self.param_groups:
            for param in group["params"]:
                param.zero_grad()

    def add_param_group(self, param_group: dict[str, Any]) -> None:
        for group in self.param_groups:
            if set(group["params"]) & set(param_group["params"]):
                raise ValueError(
                    "Some parameters appear in more than one parameter group."
                )
        self.param_groups.append(param_group)

    def state_dict(self) -> _StateDict:
        return {
            "state": copy.deepcopy(self.state),
            "param_groups": copy.deepcopy(self.param_groups),
        }

    def load_state_dict(self, state_dict: _StateDict) -> None:
        self.state = defaultdict(dict, copy.deepcopy(state_dict["state"]))
        self.param_groups = copy.deepcopy(state_dict["param_groups"])

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.param_groups})"



from typing import Dict

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid import register_model
from lucid._tensor import Tensor
from lucid.types import _Scalar


__all__ = [
    "MobileNet",
    "MobileNet_V2",
    "MobileNet_V3",
    "MobileNet_V4",
    "mobilenet",
    "mobilenet_v2",
    "mobilenet_v3_small",
    "mobilenet_v3_large",
    "mobilenet_v4_conv_small",
    "mobilenet_v4_conv_medium",
    "mobilenet_v4_conv_large",
    "mobilenet_v4_hybrid_medium",
    "mobilenet_v4_hybrid_large",
]


class _Depthwise(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, stride: int = 1) -> None:
        super().__init__()

        self.depthwise = nn.Sequential(
            nn.Conv2d(
                in_channels,
                in_channels,
                kernel_size=3,
                stride=stride,
                padding=1,
                groups=in_channels,
                bias=False,
            ),
            nn.BatchNorm2d(in_channels),
            nn.ReLU(),
        )

        self.pointwise = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.pointwise(self.depthwise(x))


class MobileNet(nn.Module):
    def __init__(self, width_multiplier: float, num_classes: int = 1000) -> None:
        super().__init__()
        alpha = width_multiplier

        self.conv1 = nn.ConvBNReLU2d(
            3, int(32 * alpha), kernel_size=3, stride=2, padding=1
        )
        self.conv2 = _Depthwise(int(32 * alpha), int(64 * alpha))

        self.conv3 = nn.Sequential(
            _Depthwise(int(64 * alpha), int(128 * alpha), stride=2),
            _Depthwise(int(128 * alpha), int(128 * alpha), stride=1),
        )
        self.conv4 = nn.Sequential(
            _Depthwise(int(128 * alpha), int(256 * alpha), stride=2),
            _Depthwise(int(256 * alpha), int(256 * alpha), stride=1),
        )

        self.conv5 = nn.Sequential(
            _Depthwise(int(256 * alpha), int(512 * alpha), stride=2),
            *[
                _Depthwise(int(512 * alpha), int(512 * alpha), stride=1)
                for _ in range(5)
            ],
        )
        self.conv6 = _Depthwise(int(512 * alpha), int(1024 * alpha), stride=2)
        self.conv7 = _Depthwise(int(1024 * alpha), int(1024 * alpha), stride=2)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(int(1024 * alpha), num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = self.conv5(x)
        x = self.conv6(x)
        x = self.conv7(x)

        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)

        return x


class _InvertedBottleneck(nn.Module):
    def __init__(
        self, in_channels: int, out_channels: int, t: int, stride: int = 1
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride

        expand = nn.Sequential(
            nn.Conv2d(in_channels, in_channels * t, kernel_size=1, bias=False),
            nn.BatchNorm2d(in_channels * t),
            nn.ReLU6(),
        )
        depthwise = nn.Sequential(
            nn.Conv2d(
                in_channels * t,
                in_channels * t,
                kernel_size=3,
                stride=stride,
                padding=1,
                groups=in_channels * t,
                bias=False,
            ),
            nn.BatchNorm2d(in_channels * t),
            nn.ReLU6(),
        )
        pointwise = nn.Sequential(
            nn.Conv2d(in_channels * t, out_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_channels),
        )

        residual_list = []
        if t > 1:
            residual_list += [expand]
        residual_list += [depthwise, pointwise]

        self.residual = nn.Sequential(*residual_list)

    def forward(self, x: Tensor) -> Tensor:
        if self.stride == 1 and self.in_channels == self.out_channels:
            out = self.residual(x) + x
        else:
            out = self.residual(x)
        return out


class MobileNet_V2(nn.Module):
    def __init__(self, num_classes: int = 1000) -> None:
        super().__init__()

        self.conv_first = nn.Sequential(
            nn.Conv2d(3, 32, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(32),
            nn.ReLU6(),
        )

        self.bottlenecks = nn.Sequential(
            self._make_stage(32, 16, t=1, n=1),
            self._make_stage(16, 24, t=6, n=2, stride=2),
            self._make_stage(24, 32, t=6, n=3, stride=2),
            self._make_stage(32, 64, t=6, n=4, stride=2),
            self._make_stage(64, 96, t=6, n=3),
            self._make_stage(96, 160, t=6, n=3, stride=2),
            self._make_stage(160, 320, t=6, n=1),
        )

        self.conv_last = nn.Sequential(
            nn.Conv2d(320, 1280, kernel_size=1, bias=False),
            nn.BatchNorm2d(1280),
            nn.ReLU6(),
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Sequential(nn.Dropout(p=0.2), nn.Linear(1280, num_classes))

    def _make_stage(
        self, in_channels: int, out_channels: int, t: int, n: int, stride: int = 1
    ) -> nn.Sequential:
        layers = [_InvertedBottleneck(in_channels, out_channels, t, stride=stride)]
        in_channels = out_channels
        for _ in range(n - 1):
            layers.append(_InvertedBottleneck(in_channels, out_channels, t, stride=1))

        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv_first(x)
        x = self.bottlenecks(x)
        x = self.conv_last(x)

        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)

        return x


class _SEBlock(nn.Module):
    def __init__(self, in_channels: int, reduction: int = 4) -> None:
        super().__init__()
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        self.fc1 = nn.Linear(in_channels, in_channels // reduction)
        self.fc2 = nn.Linear(in_channels // reduction, in_channels)

        self.relu = nn.ReLU()
        self.hsigmoid = nn.HardSigmoid()

    def forward(self, x: Tensor) -> Tensor:
        y = self.avgpool(x).squeeze(axis=(-1, -2))
        y = self.relu(self.fc1(y))
        y = self.hsigmoid(self.fc2(y))

        y = y.unsqueeze(axis=(-1, -2))
        out = x * y
        return out


class _InvertedBottleneck_V3(nn.Module):
    def __init__(
        self,
        in_channels: int,
        mid_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        use_se: bool,
        use_hswish: bool,
    ) -> None:
        super().__init__()
        self.do_skip = stride == 1 and in_channels == out_channels

        expand = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(mid_channels, momentum=0.99),
            nn.HardSwish() if use_hswish else nn.ReLU(),
        )

        depthwise = nn.Sequential(
            nn.Conv2d(
                mid_channels,
                mid_channels,
                kernel_size,
                stride,
                padding="same",
                groups=mid_channels,
                bias=False,
            ),
            nn.BatchNorm2d(mid_channels, momentum=0.99),
            nn.HardSwish() if use_hswish else nn.ReLU(),
        )

        se_block = _SEBlock(mid_channels) if use_se else None

        pointwise = nn.Sequential(
            nn.Conv2d(mid_channels, out_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_channels, momentum=0.99),
        )

        layers = []
        if in_channels < mid_channels:
            layers.append(expand)
        layers.append(depthwise)

        if se_block is not None:
            layers.append(se_block)
        layers.append(pointwise)

        self.residual = nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        if self.do_skip:
            out = self.residual(x) + x
        else:
            out = self.residual(x)
        return out


class MobileNet_V3(nn.Module):
    def __init__(
        self, bottleneck_cfg: list, last_channels: int, num_classes: int = 1000
    ) -> None:
        super().__init__()

        self.conv_first = nn.Sequential(
            nn.Conv2d(3, 16, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(16),
            nn.HardSwish(),
        )

        in_channels = 16
        bottleneck_layers = []
        for (
            kernel_size,
            mid_channels,
            out_channels,
            use_se,
            use_hswish,
            stride,
        ) in bottleneck_cfg:
            bottleneck_layers.append(
                _InvertedBottleneck_V3(
                    in_channels,
                    mid_channels,
                    out_channels,
                    kernel_size,
                    stride,
                    use_se,
                    use_hswish,
                )
            )
            in_channels = out_channels
        self.bottlenecks = nn.Sequential(*bottleneck_layers)

        self.conv_last = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.HardSwish(),
        )

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc1 = nn.Sequential(
            nn.Linear(mid_channels, last_channels),
            nn.HardSwish(),
        )
        self.fc2 = nn.Sequential(
            nn.Dropout(p=0.2), nn.Linear(last_channels, num_classes)
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv_first(x)
        x = self.bottlenecks(x)
        x = self.conv_last(x)

        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)
        x = self.fc1(x)
        x = self.fc2(x)

        return x


def _make_divisible(
    value: float,
    divisor: int,
    min_value: float | None = None,
    round_down_protect: bool = True,
) -> int:
    if min_value is None:
        min_value = divisor

    new_value = max(min_value, int(value + divisor / 2) // divisor * divisor)
    if round_down_protect and new_value < 0.9 * value:
        new_value += divisor

    return int(new_value)


def _make_conv_block(
    in_channels: int,
    out_channels: int,
    kernel_size: int = 3,
    stride: int = 1,
    groups: int = 1,
    bias: bool = False,
    norm: bool = True,
    act: bool = True,
) -> nn.Sequential:
    conv = nn.Sequential()
    conv.add_module(
        "conv",
        nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding="same",
            bias=bias,
            groups=groups,
        ),
    )
    if norm:
        conv.add_module("bn", nn.BatchNorm2d(out_channels))
    if act:
        conv.add_module("act", nn.ReLU6())

    return conv


class _InvertedResidual(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int,
        expand_ratio: int,
        act: bool = False,
        se: bool = False,
    ) -> None:
        super().__init__()
        assert stride in [1, 2]
        self.stride = stride

        hid_channels = int(round(in_channels * expand_ratio))
        self.block = nn.Sequential()

        if expand_ratio != 1:
            self.block.add_module(
                "exp_1x1",
                _make_conv_block(
                    in_channels, hid_channels, kernel_size=3, stride=stride
                ),
            )
        if se:
            self.block.add_module(
                "conv_3x3",
                _make_conv_block(
                    hid_channels,
                    hid_channels,
                    kernel_size=3,
                    stride=stride,
                    groups=hid_channels,
                ),
            )

        self.block.add_module(
            "red_1x1",
            _make_conv_block(
                hid_channels, out_channels, kernel_size=1, stride=1, act=act
            ),
        )
        self.use_residual = self.stride == 1 and in_channels == out_channels

    def forward(self, x: Tensor) -> Tensor:
        if self.use_residual:
            return x + self.block(x)

        return self.block(x)


class _UniversalInvertedBottleneck(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        start_dw_kernel_size: int,
        mid_dw_kernel_size: int,
        mid_dw_downsample: bool,
        stride: int,
        expand_ratio: int,
    ) -> None:
        super().__init__()
        self.start_dw_kernel_size = start_dw_kernel_size
        if self.start_dw_kernel_size:
            stride_ = stride if not mid_dw_downsample else 1
            self.start_dw = _make_conv_block(
                in_channels,
                in_channels,
                kernel_size=start_dw_kernel_size,
                stride=stride_,
                groups=in_channels,
                act=False,
            )

        expand_filters = _make_divisible(in_channels * expand_ratio, divisor=8)
        self.expand_conv = _make_conv_block(in_channels, expand_filters, kernel_size=1)

        self.mid_dw_kernel_size = mid_dw_kernel_size
        if self.mid_dw_kernel_size:
            stride_ = stride if mid_dw_downsample else 1
            self.middle_dw = _make_conv_block(
                expand_filters,
                expand_filters,
                kernel_size=mid_dw_kernel_size,
                stride=stride_,
                groups=expand_filters,
            )

        self.proj_conv = _make_conv_block(
            expand_filters, out_channels, kernel_size=1, act=False
        )

    def forward(self, x: Tensor) -> Tensor:
        if self.start_dw_kernel_size:
            x = self.start_dw(x)

        x = self.expand_conv(x)
        if self.mid_dw_kernel_size:
            x = self.middle_dw(x)

        x = self.proj_conv(x)
        return x


class _MultiQueryAttention(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_heads: int,
        key_dim: int,
        value_dim: int,
        query_h_stride: int,
        query_w_stride: int,
        kv_stride: int,
        dw_kernel_size: int = 3,
        dropout: float = 0.0,
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.key_dim = key_dim
        self.value_dim = value_dim
        self.query_h_stride = query_h_stride
        self.query_w_stride = query_w_stride
        self.kv_stride = kv_stride
        self.dw_kernel_size = dw_kernel_size
        self.dropout = dropout

        self.head_dim = key_dim // num_heads

        if self.query_h_stride > 1 or query_w_stride > 1:
            self.query_downsample_norm = nn.BatchNorm2d(in_channels)
        self.query_proj = _make_conv_block(
            in_channels, num_heads * key_dim, kernel_size=1, norm=False, act=False
        )

        if self.kv_stride > 1:
            self.key_dw_conv = _make_conv_block(
                in_channels,
                in_channels,
                kernel_size=dw_kernel_size,
                stride=kv_stride,
                groups=in_channels,
                act=False,
            )
            self.value_dw_conv = _make_conv_block(
                in_channels,
                in_channels,
                kernel_size=dw_kernel_size,
                stride=kv_stride,
                groups=in_channels,
                act=False,
            )

        self.key_proj = _make_conv_block(
            in_channels, key_dim, kernel_size=1, norm=False, act=False
        )
        self.value_proj = _make_conv_block(
            in_channels, key_dim, kernel_size=1, norm=False, act=False
        )
        self.output_proj = _make_conv_block(
            num_heads * key_dim, in_channels, kernel_size=1, norm=False, act=False
        )
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x: Tensor) -> Tensor:
        N = x.shape[0]
        if self.query_h_stride > 1 or self.query_w_stride > 1:
            q = F.avg_pool2d(
                x,
                kernel_size=(self.query_h_stride, self.query_w_stride),
                stride=(self.query_h_stride, self.query_w_stride),
            )
            q = self.query_downsample_norm(q)
            q = self.query_proj(q)
        else:
            q = self.query_proj(x)

        px = q.shape[2]
        q = q.reshape(N, self.num_heads, -1, self.key_dim)

        if self.kv_stride > 1:
            k = self.key_dw_conv(x)
            k = self.key_proj(k)
            v = self.value_dw_conv(x)
            v = self.value_proj(v)
        else:
            k = self.key_proj(x)
            v = self.value_proj(x)

        k = k.reshape(N, 1, self.key_dim, -1)
        v = v.reshape(N, 1, -1, self.key_dim)

        attn_score = (q @ k) / (self.head_dim**0.5)
        attn_score = self.dropout(attn_score)
        attn_score = F.softmax(attn_score, axis=-1)

        context = attn_score @ v
        context = context.reshape(N, self.num_heads * self.key_dim, px, px)

        out = self.output_proj(context)
        return out


class _LayerScale(nn.Module):
    def __init__(self, in_channels: int, init_value: _Scalar) -> None:
        super().__init__()
        self.init_value = init_value
        self.gamma = nn.Parameter(self.init_value * lucid.ones(in_channels, 1, 1))

    def forward(self, x: Tensor) -> Tensor:
        return x * self.gamma


class _MultiHeadSelfAttention(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_heads: int,
        key_dim: int,
        value_dim: int,
        query_h_stride: int,
        query_w_stride: int,
        kv_stride: int,
        use_layer_scale: bool,
        use_multi_query: bool,
        use_residual: bool = True,
    ) -> None:
        super().__init__()
        self.query_h_stride = query_h_stride
        self.query_w_stride = query_w_stride
        self.kv_stride = kv_stride
        self.use_layer_scale = use_layer_scale
        self.use_multi_query = use_multi_query
        self.use_residual = use_residual

        self.input_norm = nn.BatchNorm2d(in_channels)
        if self.use_multi_query:
            self.multi_query_attention = _MultiQueryAttention(
                in_channels,
                num_heads,
                key_dim,
                value_dim,
                query_h_stride,
                query_w_stride,
                kv_stride,
            )
        else:
            self.multi_head_attention = nn.MultiHeadAttention(
                in_channels, num_heads, kdim=key_dim
            )

        if self.use_layer_scale:
            self.layer_scale_init_value = 1e-5
            self.layer_scale = _LayerScale(in_channels, self.layer_scale_init_value)

    def forward(self, x: Tensor) -> Tensor:
        shortcut = x
        x = self.input_norm(x)

        if self.use_multi_query:
            x = self.multi_query_attention(x)
        else:
            x = self.multi_head_attention(x, x, x)

        if self.use_layer_scale:
            x = self.layer_scale(x)
        if self.use_residual:
            x += shortcut

        return x


def build_v4_blocks(layer_spec: dict) -> nn.Sequential:
    if not layer_spec.get("block_name"):
        return nn.Sequential()

    block_names = layer_spec["block_name"]
    layers = nn.Sequential()

    if block_names == "convbn":
        schema = ["in_channels", "out_channels", "kernel_size", "stride"]
        for i in range(layer_spec["num_blocks"]):
            args = dict(zip(schema, layer_spec["block_specs"][i]))
            layers.add_module(f"convbn_{i}", _make_conv_block(**args))

    elif block_names == "uib":
        schema = [
            "in_channels",
            "out_channels",
            "start_dw_kernel_size",
            "mid_dw_kernel_size",
            "mid_dw_downsample",
            "stride",
            "expand_ratio",
            "mha",
        ]
        for i in range(layer_spec["num_blocks"]):
            args = dict(zip(schema, layer_spec["block_specs"][i]))
            mha = args.pop("mha") if "mha" in args else 0

            layers.add_module(f"uib_{i}", _UniversalInvertedBottleneck(**args))
            if mha:
                mha_schema = [
                    "in_channels",
                    "num_heads",
                    "key_dim",
                    "value_dim",
                    "query_h_stride",
                    "query_w_stride",
                    "kv_stride",
                    "use_layer_scale",
                    "use_multi_query",
                    "use_residual",
                ]
                args = dict(zip(mha_schema, [args["out_channels"]] + mha))
                layers.add_module(f"mha_{i}", _MultiHeadSelfAttention(**args))

    elif block_names == "fused_ib":
        schema = ["in_channels", "out_channels", "stride", "expand_ratio", "act"]
        for i in range(layer_spec["num_blocks"]):
            args = dict(zip(schema, layer_spec["block_specs"][i]))
            layers.add_module(f"fused_ib_{i}", _InvertedResidual(**args))

    else:
        raise NotImplementedError

    return layers


class MobileNet_V4(nn.Module):
    def __init__(self, cfg: Dict[str, dict], num_classes: int = 1000) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.conv0 = build_v4_blocks(cfg["conv0"])

        self.layers = nn.ModuleList()
        for i in range(1, 6):
            self.layers.append(build_v4_blocks(cfg[f"layer{i}"]))

        last_channels = cfg["layer5"]["block_specs"][-1][1]
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(last_channels, num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv0(x)
        for layer in self.layers:
            x = layer(x)

        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)

        return x


@register_model
def mobilenet(
    width_multiplier: float = 1.0, num_classes: int = 1000, **kwargs
) -> MobileNet:
    return MobileNet(width_multiplier, num_classes, **kwargs)


@register_model
def mobilenet_v2(num_classes: int = 1000, **kwargs) -> MobileNet_V2:
    return MobileNet_V2(num_classes, **kwargs)


@register_model
def mobilenet_v3_small(num_classes: int = 1000, **kwargs) -> MobileNet_V3:
    cfg = [
        [3, 16, 16, True, False, 2],
        [3, 72, 24, False, False, 2],
        [3, 88, 24, False, False, 1],
        [5, 96, 40, True, True, 2],
        [5, 240, 40, True, True, 1],
        [5, 240, 40, True, True, 1],
        [5, 120, 48, True, True, 1],
        [5, 144, 48, True, True, 1],
        [5, 288, 96, True, True, 2],
        [5, 576, 96, True, True, 1],
        [5, 576, 96, True, True, 1],
    ]
    return MobileNet_V3(
        bottleneck_cfg=cfg, last_channels=1024, num_classes=num_classes, **kwargs
    )


@register_model
def mobilenet_v3_large(num_classes: int = 1000, **kwargs) -> MobileNet_V3:
    cfg = [
        [3, 16, 16, False, False, 1],
        [3, 64, 24, False, False, 2],
        [3, 72, 24, False, False, 1],
        [5, 72, 40, True, False, 2],
        [5, 120, 40, True, False, 1],
        [5, 120, 40, True, False, 1],
        [3, 240, 80, False, True, 2],
        [3, 200, 80, False, True, 1],
        [3, 184, 80, False, True, 1],
        [3, 184, 80, False, True, 1],
        [3, 480, 112, True, True, 1],
        [3, 672, 112, True, True, 1],
        [5, 672, 160, True, True, 2],
        [5, 960, 160, True, True, 1],
        [5, 960, 160, True, True, 1],
    ]
    return MobileNet_V3(
        bottleneck_cfg=cfg, last_channels=1280, num_classes=num_classes, **kwargs
    )


@register_model
def mobilenet_v4_conv_small(num_classes: int = 1000, **kwargs) -> MobileNet_V4:
    cfg = dict(
        conv0={
            "block_name": "convbn",
            "num_blocks": 1,
            "block_specs": [[3, 32, 3, 2]],
        },
        layer1={
            "block_name": "convbn",
            "num_blocks": 2,
            "block_specs": [[32, 32, 3, 2], [32, 32, 1, 1]],
        },
        layer2={
            "block_name": "convbn",
            "num_blocks": 2,
            "block_specs": [[32, 96, 3, 2], [96, 64, 1, 1]],
        },
        layer3={
            "block_name": "uib",
            "num_blocks": 6,
            "block_specs": [
                [64, 96, 5, 5, True, 2, 3],
                [96, 96, 0, 3, True, 1, 2],
                [96, 96, 0, 3, True, 1, 2],
                [96, 96, 0, 3, True, 1, 2],
                [96, 96, 0, 3, True, 1, 2],
                [96, 96, 3, 0, True, 1, 4],
            ],
        },
        layer4={
            "block_name": "uib",
            "num_blocks": 6,
            "block_specs": [
                [96, 128, 3, 3, True, 2, 6],
                [128, 128, 5, 5, True, 1, 4],
                [128, 128, 0, 5, True, 1, 4],
                [128, 128, 0, 5, True, 1, 3],
                [128, 128, 0, 3, True, 1, 4],
                [128, 128, 0, 3, True, 1, 4],
            ],
        },
        layer5={
            "block_name": "convbn",
            "num_blocks": 2,
            "block_specs": [[128, 960, 1, 1], [960, 1280, 1, 1]],
        },
    )
    return MobileNet_V4(cfg, num_classes, **kwargs)


@register_model
def mobilenet_v4_conv_medium(num_classes: int = 1000, **kwargs) -> MobileNet_V4:
    cfg = dict(
        conv0={
            "block_name": "convbn",
            "num_blocks": 1,
            "block_specs": [[3, 32, 3, 2]],
        },
        layer1={
            "block_name": "fused_ib",
            "num_blocks": 1,
            "block_specs": [[32, 48, 2, 4.0, True]],
        },
        layer2={
            "block_name": "uib",
            "num_blocks": 2,
            "block_specs": [[48, 80, 3, 5, True, 2, 4], [80, 80, 3, 3, True, 1, 2]],
        },
        layer3={
            "block_name": "uib",
            "num_blocks": 8,
            "block_specs": [
                [80, 160, 3, 5, True, 2, 6],
                [160, 160, 3, 3, True, 1, 4],
                [160, 160, 3, 3, True, 1, 4],
                [160, 160, 3, 5, True, 1, 4],
                [160, 160, 3, 3, True, 1, 4],
                [160, 160, 3, 0, True, 1, 4],
                [160, 160, 0, 0, True, 1, 2],
                [160, 160, 3, 0, True, 1, 4],
            ],
        },
        layer4={
            "block_name": "uib",
            "num_blocks": 11,
            "block_specs": [
                [160, 256, 5, 5, True, 2, 6],
                [256, 256, 5, 5, True, 1, 4],
                [256, 256, 3, 5, True, 1, 4],
                [256, 256, 3, 5, True, 1, 4],
                [256, 256, 0, 0, True, 1, 4],
                [256, 256, 3, 0, True, 1, 4],
                [256, 256, 3, 5, True, 1, 2],
                [256, 256, 5, 5, True, 1, 4],
                [256, 256, 0, 0, True, 1, 4],
                [256, 256, 0, 0, True, 1, 4],
                [256, 256, 5, 0, True, 1, 2],
            ],
        },
        layer5={
            "block_name": "convbn",
            "num_blocks": 2,
            "block_specs": [[256, 960, 1, 1], [960, 1280, 1, 1]],
        },
    )
    return MobileNet_V4(cfg, num_classes, **kwargs)


@register_model
def mobilenet_v4_conv_large(num_classes: int = 1000, **kwargs) -> MobileNet_V4:
    cfg = dict(
        conv0={
            "block_name": "convbn",
            "num_blocks": 1,
            "block_specs": [[3, 24, 3, 2]],
        },
        layer1={
            "block_name": "fused_ib",
            "num_blocks": 1,
            "block_specs": [[24, 48, 2, 4.0, True]],
        },
        layer2={
            "block_name": "uib",
            "num_blocks": 2,
            "block_specs": [[48, 96, 3, 5, True, 2, 4], [96, 96, 3, 3, True, 1, 4]],
        },
        layer3={
            "block_name": "uib",
            "num_blocks": 11,
            "block_specs": [
                [96, 192, 3, 5, True, 2, 4],
                [192, 192, 3, 3, True, 1, 4],
                [192, 192, 3, 3, True, 1, 4],
                [192, 192, 3, 3, True, 1, 4],
                [192, 192, 3, 5, True, 1, 4],
                [192, 192, 5, 3, True, 1, 4],
                [192, 192, 5, 3, True, 1, 4],
                [192, 192, 5, 3, True, 1, 4],
                [192, 192, 5, 3, True, 1, 4],
                [192, 192, 5, 3, True, 1, 4],
                [192, 192, 3, 0, True, 1, 4],
            ],
        },
        layer4={
            "block_name": "uib",
            "num_blocks": 13,
            "block_specs": [
                [192, 512, 5, 5, True, 2, 4],
                [512, 512, 5, 5, True, 1, 4],
                [512, 512, 5, 5, True, 1, 4],
                [512, 512, 5, 5, True, 1, 4],
                [512, 512, 5, 0, True, 1, 4],
                [512, 512, 5, 3, True, 1, 4],
                [512, 512, 5, 0, True, 1, 4],
                [512, 512, 5, 0, True, 1, 4],
                [512, 512, 5, 3, True, 1, 4],
                [512, 512, 5, 5, True, 1, 4],
                [512, 512, 5, 0, True, 1, 4],
                [512, 512, 5, 0, True, 1, 4],
                [512, 512, 5, 0, True, 1, 4],
            ],
        },
        layer5={
            "block_name": "convbn",
            "num_blocks": 2,
            "block_specs": [[512, 960, 1, 1], [960, 1280, 1, 1]],
        },
    )
    return MobileNet_V4(cfg, num_classes, **kwargs)


def _mha_specs(num_heads: int, key_dim: int, value_dim: int, px: int) -> list[int]:
    if px == 24:
        kv_stride = 2
    elif px == 12:
        kv_stride = 1
    query_h_stride = query_w_stride = 1
    use_layer_scale = use_multi_query = use_residual = True
    return [
        num_heads,
        key_dim,
        value_dim,
        query_h_stride,
        query_w_stride,
        kv_stride,
        use_layer_scale,
        use_multi_query,
        use_residual,
    ]


@register_model
def mobilenet_v4_hybrid_medium(num_classes: int = 1000, **kwargs) -> MobileNet_V4:
    cfg = dict(
        conv0={
            "block_name": "convbn",
            "num_blocks": 1,
            "block_specs": [[3, 32, 3, 2]],
        },
        layer1={
            "block_name": "fused_ib",
            "num_blocks": 1,
            "block_specs": [[32, 48, 2, 4.0, True]],
        },
        layer2={
            "block_name": "uib",
            "num_blocks": 2,
            "block_specs": [[48, 80, 3, 5, True, 2, 4], [80, 80, 3, 3, True, 1, 2]],
        },
        layer3={
            "block_name": "uib",
            "num_blocks": 8,
            "block_specs": [
                [80, 160, 3, 5, True, 2, 6],
                [160, 160, 0, 0, True, 1, 2],
                [160, 160, 3, 3, True, 1, 4],
                [160, 160, 3, 5, True, 1, 4, _mha_specs(4, 64, 64, 24)],
                [160, 160, 3, 3, True, 1, 4, _mha_specs(4, 64, 64, 24)],
                [160, 160, 3, 0, True, 1, 4, _mha_specs(4, 64, 64, 24)],
                [160, 160, 3, 3, True, 1, 4, _mha_specs(4, 64, 64, 24)],
                [160, 160, 3, 0, True, 1, 4],
            ],
        },
        layer4={
            "block_name": "uib",
            "num_blocks": 12,
            "block_specs": [
                [160, 256, 5, 5, True, 2, 6],
                [256, 256, 5, 5, True, 1, 4],
                [256, 256, 3, 5, True, 1, 4],
                [256, 256, 3, 5, True, 1, 4],
                [256, 256, 0, 0, True, 1, 2],
                [256, 256, 3, 5, True, 1, 2],
                [256, 256, 0, 0, True, 1, 2],
                [256, 256, 0, 0, True, 1, 4, _mha_specs(4, 64, 64, 12)],
                [256, 256, 3, 0, True, 1, 4, _mha_specs(4, 64, 64, 12)],
                [256, 256, 5, 5, True, 1, 4, _mha_specs(4, 64, 64, 12)],
                [256, 256, 5, 0, True, 1, 4, _mha_specs(4, 64, 64, 12)],
                [256, 256, 5, 0, True, 1, 4],
            ],
        },
        layer5={
            "block_name": "convbn",
            "num_blocks": 2,
            "block_specs": [[256, 960, 1, 1], [960, 1280, 1, 1]],
        },
    )
    return MobileNet_V4(cfg, num_classes, **kwargs)


@register_model
def mobilenet_v4_hybrid_large(num_classes: int = 1000, **kwargs) -> MobileNet_V4:
    cfg = dict(
        conv0={"block_name": "convbn", "num_blocks": 1, "block_specs": [[3, 24, 3, 2]]},
        layer1={
            "block_name": "fused_ib",
            "num_blocks": 1,
            "block_specs": [[24, 48, 2, 4.0, True]],
        },
        layer2={
            "block_name": "uib",
            "num_blocks": 2,
            "block_specs": [[48, 96, 3, 5, True, 2, 4], [96, 96, 3, 3, True, 1, 4]],
        },
        layer3={
            "block_name": "uib",
            "num_blocks": 11,
            "block_specs": [
                [96, 192, 3, 5, True, 2, 4],
                [192, 192, 3, 3, True, 1, 4],
                [192, 192, 3, 3, True, 1, 4],
                [192, 192, 3, 3, True, 1, 4],
                [192, 192, 3, 5, True, 1, 4],
                [192, 192, 5, 3, True, 1, 4],
                [192, 192, 5, 3, True, 1, 4, _mha_specs(8, 48, 48, 24)],
                [192, 192, 5, 3, True, 1, 4, _mha_specs(8, 48, 48, 24)],
                [192, 192, 5, 3, True, 1, 4, _mha_specs(8, 48, 48, 24)],
                [192, 192, 5, 3, True, 1, 4, _mha_specs(8, 48, 48, 24)],
                [192, 192, 3, 0, True, 1, 4],
            ],
        },
        layer4={
            "block_name": "uib",
            "num_blocks": 14,
            "block_specs": [
                [192, 512, 5, 5, True, 2, 4],
                [512, 512, 5, 5, True, 1, 4],
                [512, 512, 5, 5, True, 1, 4],
                [512, 512, 5, 5, True, 1, 4],
                [512, 512, 5, 0, True, 1, 4],
                [512, 512, 5, 3, True, 1, 4],
                [512, 512, 5, 0, True, 1, 4],
                [512, 512, 5, 0, True, 1, 4],
                [512, 512, 5, 3, True, 1, 4],
                [512, 512, 5, 5, True, 1, 4, _mha_specs(8, 64, 64, 12)],
                [512, 512, 5, 0, True, 1, 4, _mha_specs(8, 64, 64, 12)],
                [512, 512, 5, 0, True, 1, 4, _mha_specs(8, 64, 64, 12)],
                [512, 512, 5, 0, True, 1, 4, _mha_specs(8, 64, 64, 12)],
                [512, 512, 5, 0, True, 1, 4],
            ],
        },
        layer5={
            "block_name": "convbn",
            "num_blocks": 2,
            "block_specs": [[512, 960, 1, 1], [960, 1280, 1, 1]],
        },
    )
    return MobileNet_V4(cfg, num_classes, **kwargs)
